package com.isencia.metamodeltest;


import java.lang.reflect.Field;
import java.lang.reflect.Modifier;

import javax.persistence.EntityManagerFactory;
import javax.persistence.Persistence;
import javax.persistence.metamodel.EntityType;

/**
 * <code>MetamodelTest</code> demonstrates a problem with @ManyToOne attributes when using an IBM JDK.
 * It has been detected (in production applications using EclipseLink JPA 2 with OSGI),
 * that some @ManyToOne attributes sometimes are not represented in the JPA meta-model,
 * although their mapping is perfectly valid, and they are always correctly represented
 * when using a Sun JDK, and they mostly also are so with the IBM JDK.
 * <br>
 * I have not found any reliable pattern for the failure, except that the missing attributes
 * always are @ManyToOne associations.
 * <br>
 * To exclude the OSGI platform or weaving as possible causes, I have set up a simple java executable test.
 * The code below loops eternally, creating each time a new EntityManagerFactory to force (re)deploy,
 * and checking all registered entities for any missing attributes. Whenever a missing attribute is detected,
 * the EL exception is logged and the loop is interrupted.
 * <br>
 * I have defined a persistenceUnit with 3 simple entities, simulating 2 levels of association nesting:
 * <ul>
 * <li>Test (the base entity)</li> 
 * <li>TestContainer (an entity that contains references to Test)</li>
 * <li>TestContainerContainer (an entity that contains references to TestContainer)</li>
 * </ul>
 * <br>
 * If running this code with a Sun JDK (5 or higher), the application never ends as deploy never fails.
 * If running this code with an IBM JDK (5 or higher), the application ends after a random number of loops,
 * as the deploy seems to fail and an attribute is detected to be missing in the JPA meta-model.
 * 
 * @author Jan Vermeulen (jan.vermeulen@isencia.com)
 *
 */
public class MetamodelTest {
	public static void main(String[] args) {
		int i = 0;
		while (true) {
			System.out.println("Deploying test... (counter: "+(++i)+")");
			EntityManagerFactory factory = null;
			
			try {
				// (re)deploy the EntityManagerFactory
				factory = Persistence.createEntityManagerFactory("test");
			
				// check for all registered entities if all member fields of the class
				// have a corresponding attribute in the meta-model
				for (EntityType<?> entity : factory.getMetamodel().getEntities())
					for (Field field : entity.getJavaType().getDeclaredFields()) {
						// ignore static and woven fields 
						if (Modifier.isStatic(field.getModifiers()) || field.getName().charAt(0) == '_')
							continue;
						entity.getAttribute(field.getName());
					}
			} catch (Exception e) {
				// report the missing attribute and get out
				System.out.println(e.getMessage());
				break;
			} finally {
				if (factory != null)
					factory.close();
			}
		}
	}
}
